# This test code was written by the `hypothesis.extra.ghostwriter` module
# and is provided under the Creative Commons Zero public domain dedication.
import typing

import pytest
from hypothesis import given
from hypothesis import strategies as st

from grottext.ha.interface import FakeConf
from grottext.ha.mqtt import cleanup_mqtt_values_field, is_valid_mqtt_topic, make_payload


@given(key_name=st.text())
def test_fuzz_is_valid_mqtt_topic(key_name: str) -> None:
    is_valid_mqtt_topic(key_name=key_name)


@given(
    conf=st.builds(FakeConf),
    device=st.text(),
    key=st.text(),
    name=st.one_of(st.none(), st.text()),
)
def test_fuzz_make_payload(conf: FakeConf, device: str, key: str, name: typing.Optional[str]) -> None:
    with pytest.raises(AttributeError):
        make_payload(conf=conf, device=device, key=key, name=name)


@given(
    values=st.dictionaries(
        st.text(),
        st.integers() | st.text(),
    )
)
def test_fuzz_cleanup_values(values: typing.Dict[str, typing.Any]) -> None:
    cleanup_mqtt_values_field(values=values)



import json
import traceback
from datetime import datetime, timezone
from typing import Any, Dict, cast

from grottext.ha.__version__ import __version__
from grottext.ha.constants import (
    CONFIG_TOPIC,
    MQTT_HOST_CONF_KEY,
    MQTT_PORT_CONF_KEY,
    MQTT_RETAIN_CONF_KEY,
    STATE_TOPIC,
)
from grottext.ha.interface import FakeConf
from grottext.ha.mqtt import (
    cleanup_mqtt_values_field,
    is_valid_mqtt_topic,
    make_payload,
    publish_multiple,
    publish_single,
)

__pv_config: Dict[str, bool] = {}


def grottext(conf: FakeConf, data: str, jsonmsg: str) -> int:
    """Allow pushing to HA MQTT bus, with auto discovery

    Use the plugin interface of grott to grab the raw JSON and push it to HA MQTT bus.

    How does it work:
        - Parse the raw json
        - Discard if it's a buffered message
        - Cleanup the values field (spaces, invalid characters)
        - Check if the device is already configured
        - If not, create a new sensor for each value
        - Push the values to the topic `homeassistant/grott/{device}/state`

    Parameters:
        conf: The configuration object from grott
        data: The raw data
        jsonmsg: The raw json message
    """

    required_params = [
        MQTT_HOST_CONF_KEY,
        MQTT_PORT_CONF_KEY,
    ]
    if not all(param in conf.extvar for param in required_params):
        print("Missing configuration for ha_mqtt")
        return 1

    # Need to decode the json string
    payload = cast(Dict[str, Any], json.loads(jsonmsg))

    if payload.get("buffered") == "yes":
        # Skip buffered message, HA don't support them
        if conf.verbose:
            print("\t - Grott HA - skipped buffered")
        return 5

    device_serial = payload["device"]
    values = cleanup_mqtt_values_field(payload["values"])

    # Send the last push in UTC with TZ
    dt = datetime.now(timezone.utc)
    # Add a new value to the existing values
    values["grott_last_push"] = dt.isoformat()

    # Layout can be undefined
    if not __pv_config.get(device_serial, False) and getattr(conf, "layout", None):
        configs_payloads = []
        print(f"\tGrott HA {__version__} - creating {device_serial} config in HA, {len(values.keys())} to push")
        for key in values:
            # Prevent creating invalid MQTT topics
            if not is_valid_mqtt_topic(key):
                if conf.verbose:
                    print(f"\t[Grott HA] {__version__} skipped key: {key} invalid name")
                continue
            # Generate a configuration payload
            try:
                payload = make_payload(conf, device_serial, key)
            except AttributeError as e:
                print(f"\t[Grott HA] {__version__} error while generating key: {key}, error: {e}")
            if not payload:
                print(f"\t[Grott HA] {__version__} skipped key: {key}")
                continue

            try:
                topic = CONFIG_TOPIC.format(
                    sensor_type="sensor",
                    device=device_serial,
                    attribut=key,
                )
                configs_payloads.append(
                    {
                        "topic": topic,
                        "payload": json.dumps(payload),
                        "retain": True,
                        "qos": 1,
                    }
                )
            except Exception as e:
                print(f"\t - [grott HA] {__version__} Exception while creating new sensor {key}: {e}")
                return 6

        # Create a virtual last_push key to allow tracking when there was the last data transmission

        try:
            key = "grott_last_push"
            payload = make_payload(conf, device_serial, key)
            topic = CONFIG_TOPIC.format(
                sensor_type="sensor",
                device=device_serial,
                attribut=key,
            )
            configs_payloads.append(
                {
                    "topic": topic,
                    "payload": json.dumps(payload),
                    "retain": True,
                    "qos": 1,
                }
            )
        except Exception as e:
            print(f"\t - [grott HA] {__version__} Exception while creating new sensor last push: {e}")
            return 4
        print(f"\tPushing {len(configs_payloads)} configurations payload to HA")
        publish_multiple(conf, configs_payloads)
        print("\tConfigurations pushed")
        # Now it's configured, no need to come back
        __pv_config[device_serial] = True

    if not __pv_config.get(device_serial, False):
        print(f"\t[Grott HA] {__version__} Can't configure device: {device_serial}")
        return 7

    # Push the values to the topic
    retain = conf.extvar.get(MQTT_RETAIN_CONF_KEY, False)
    if not isinstance(retain, bool):
        retain = False
    try:
        publish_single(
            conf,
            STATE_TOPIC.format(device=device_serial),
            json.dumps(values),
            retain=retain,
        )
    except Exception as e:
        print(f"[HA ext] - Exception while publishing - {e}")
        # Reset connection state in case of a problem
        if conf.verbose:
            traceback.print_exc()
        return 2
    return 0



# Test section
# In the same file to keep the plugin contained
import pytest

from grottext.ha.interface import FakeConf

test_serial = "NCO7410"
test_key = "pvpowerout"
test_layout = "test"


@pytest.fixture()
def fake_config() -> FakeConf:
    """Create a fake config with a test layout"""
    conf = FakeConf()
    conf.recorddict = {
        "test": {
            test_key: {
                "value": 122,
                "length": 4,
                "type": "num",
            }
        }
    }
    conf.layout = "test"
    conf.extvar = {}
    conf.verbose = True
    return conf



from grottext.ha.ha_types import to_dict
from grottext.ha.mappings import mapping
from grottext.ha.mqtt import MQTTConfigPayload, make_payload

from .conftest import test_key, test_serial


def test_generate_payload(fake_config):
    """Test that an auto-generated payload for MQTT configuration"""
    # Override the divider
    fake_config.recorddict["test"][test_key]["divide"] = 10
    payload = make_payload(fake_config, test_serial, test_key)
    print(payload)
    # The default divider for pvpowerout is 10
    assert payload["value_template"] == "{{ value_json.pvpowerout | float / 10 }}"
    assert payload["name"] == "PV Output (Actual)"
    assert payload["unique_id"] == "grott_NCO7410_pvpowerout"
    assert payload["state_class"] == "measurement"
    assert payload["device_class"] == "power"
    assert payload["unit_of_measurement"] == "W"


def test_generate_log_type_payload(fake_config):
    """Test a log type record"""
    # Override the divider
    fake_config.recorddict["test"][test_key]["type"] = "log"
    fake_config.recorddict["test"][test_key]["divider"] = "10"
    payload = make_payload(fake_config, test_serial, test_key)
    print(payload)
    # Should ignore the diveider
    assert payload["value_template"] == "{{ value_json.pvpowerout | float }}"


def test_generate_logpos_type_payload(fake_config):
    """Test a logpos type record"""
    # Override the type
    fake_config.recorddict["test"][test_key]["type"] = "logpos"
    payload = make_payload(fake_config, test_serial, test_key)
    print(payload)
    # The default divider for pvpowerout is 10
    assert payload["value_template"] == "{{ [value_json.pvpowerout | float, 0.0] | max }}"


def test_generate_logneg_type_payload(fake_config):
    """Test a logneg type record"""
    # Override the type
    fake_config.recorddict["test"][test_key]["type"] = "logneg"
    payload = make_payload(fake_config, test_serial, test_key)
    print(payload)
    # The default divider for pvpowerout is 10
    assert payload["value_template"] == "{{ [value_json.pvpowerout | float, 0.0] | min }}"


def test_generate_payload_without_divider(fake_config):
    "Test that an auto-generated payload for MQTT configuration"

    payload = make_payload(fake_config, test_serial, test_key)
    print(payload)
    # The default divider for pvpowerout is 10
    assert payload["value_template"] == "{{ value_json.pvpowerout | float / 1 }}"
    assert payload["name"] == "PV Output (Actual)"
    assert payload["unique_id"] == "grott_NCO7410_pvpowerout"
    assert payload["state_class"] == "measurement"
    assert payload["device_class"] == "power"
    assert payload["unit_of_measurement"] == "W"


def test_all_mapping_cover_config():
    "Test that all the mappings cover the attrib of the MQTTConfiguration"

    fields = MQTTConfigPayload.__dataclass_fields__.keys()
    # load all the mapping, then check the keys are matching the attributes of the MQTTConfiguration
    for key, value in mapping.items():
        # convert the mapping to a dict
        key_mapping = to_dict(value)
        # Check if the MQTTConfigPayload has the attributes
        for k in key_mapping:
            assert k in fields, f"Missing attribute for {key} - {k}"



from typing import Dict

from grottext.ha.ha_types import (
    ApparentPower,
    BaseSensor,
    BatteryChargeSensor,
    CurrentSensor,
    DiagnosticSensor,
    DurationSensor,
    EnergySensor,
    FrequencySensor,
    IncreasingEnergySensor,
    PercentSensor,
    PowerFactor,
    PowerSensor,
    ReactivePower,
    TemperatureSensor,
    VoltageSensor,
)

mapping: Dict[str, BaseSensor] = {
    "datalogserial": DiagnosticSensor("Datalogger serial"),
    "pvserial": DiagnosticSensor("serial"),
    "pv1voltage": VoltageSensor("PV1 Voltage"),
    "pv1current": CurrentSensor("PV1 Current"),
    "pv1watt": PowerSensor("PV1 Watt"),
    "pv2voltage": VoltageSensor("PV2 Voltage"),
    "pv2current": CurrentSensor("PV2 Current"),
    "pv2watt": PowerSensor("PV2 Watt"),
    "pv3voltage": VoltageSensor("PV3 Voltage"),
    "pv3current": CurrentSensor("PV3 Current"),
    "pv3watt": PowerSensor("PV3 Watt"),
    "pvpowerin": PowerSensor("PV Input (Actual)", icon="mdi:solar-power"),
    "pvpowerout": PowerSensor("PV Output (Actual)"),
    "pvfrequentie": FrequencySensor("Grid Frequency", icon="mdi:waveform"),
    "frequency": FrequencySensor("Grid Frequency", icon="mdi:waveform"),
    "line_freq": FrequencySensor("Grid Frequency", icon="mdi:waveform"),
    "outputfreq": FrequencySensor("Inverter output frequency", icon="mdi:waveform"),
    "outputvolt": VoltageSensor("Inverter output Voltage"),
    # Grid config
    "grid_volt": VoltageSensor("Grid Voltage"),
    "bus_volt": VoltageSensor("Bus Voltage"),
    "pvgridvoltage": VoltageSensor("Phase 1 voltage"),
    "pvgridvoltage2": VoltageSensor("Phase 2 voltage"),
    "pvgridvoltage3": VoltageSensor("Phase 3 voltage"),
    "pvgridcurrent": CurrentSensor("Phase 1 current"),
    "pvgridcurrent2": CurrentSensor("Phase 2 current"),
    "pvgridcurrent3": CurrentSensor("Phase 3 current"),
    "pvgridpower": PowerSensor("Phase 1 power"),
    "pvgridpower2": PowerSensor("Phase 2 power"),
    "pvgridpower3": PowerSensor("Phase 3 power"),
    # End grid
    "pvenergytoday": EnergySensor("Generated energy (Today)", icon="mdi:solar-power"),
    "epvtoday": EnergySensor("Solar energy today", icon="mdi:solar-power"),
    "epvToday": EnergySensor("Solar energy today", icon="mdi:solar-power"),
    "epv1today": EnergySensor("Solar PV1 energy today", icon="mdi:solar-power"),
    "epv2today": EnergySensor("Solar PV2 energy today", icon="mdi:solar-power"),
    "pvenergytotal": IncreasingEnergySensor("Generated energy (Total)", icon="mdi:solar-power"),
    "epvtotal": IncreasingEnergySensor("Lifetime solar energy", icon="mdi:solar-power"),
    "epv1total": IncreasingEnergySensor("Solar PV1 production (Total)", icon="mdi:solar-power"),
    "epv1tot": IncreasingEnergySensor("Solar PV1 production (Total)", icon="mdi:solar-power"),
    "epv2total": IncreasingEnergySensor("Solar PV2 production (Total)", icon="mdi:solar-power"),
    "epv2tot": IncreasingEnergySensor("Solar PV3 production (Total)", icon="mdi:solar-power"),
    # For SPH compatibility
    "epvTotal": IncreasingEnergySensor("Generated PV energy (Total)", icon="mdi:solar-power"),
    "pactouserr": PowerSensor("Import from grid", icon="mdi:transmission-tower-export"),
    "pactousertot": PowerSensor("Import from grid total", icon="mdi:transmission-tower-export"),
    "pactogridr": PowerSensor("Export to grid", icon="mdi:solar-power"),
    "pactogridtot": PowerSensor("Export to grid total", icon="mdi:solar-power"),
    "pvstatus": DiagnosticSensor("PV Status", icon="mdi:power-settings"),
    "totworktime": DurationSensor(
        "Working time",
        value_template="{{ value_json.totworktime| float / 7200 | round(2) }}",
    ),
    "pvtemperature": TemperatureSensor("Inverter temperature", icon="mdi:thermometer"),
    "pvipmtemperature": TemperatureSensor("Intelligent Power Management temperature", icon="mdi:thermometer"),
    "pvboottemperature": TemperatureSensor("Inverter boost temperature", icon="mdi:thermometer"),
    "pvboosttemp": TemperatureSensor("Inverter boost temperature", icon="mdi:thermometer"),
    # Energy
    "etogrid_tod": EnergySensor("Export to Grid Energy - Today", icon="mdi:transmission-tower-import"),
    "etogrid_tot": IncreasingEnergySensor("Export to Grid Energy - Total", icon="mdi:transmission-tower-import"),
    "etouser_tod": EnergySensor("Import from Grid Energy - Today", icon="mdi:transmission-tower-export"),
    "etouser_tot": IncreasingEnergySensor("Import from Grid Energy - Total", icon="mdi:transmission-tower-export"),
    # Need to investigate
    "elocalload_tod": EnergySensor("Load Consumption Energy - Today", icon="mdi:solar-power"),
    "elocalload_tot": IncreasingEnergySensor("Load Consumption Energy - Total", icon="mdi:solar-power"),
    "AC_InWatt": PowerSensor("Grid input power"),
    "AC_InVA": ApparentPower("Grid input apparent power"),
    "plocaloadr": PowerSensor("Local load consumption", icon="mdi:transmission-tower-export"),
    # extension data
    "grott_last_push": BaseSensor(
        "Grott last data push",
        device_class="timestamp",
        value_template="{{value_json.grott_last_push}}",
    ),
    "grott_last_measure": BaseSensor(
        "Grott last measure",
        device_class="timestamp",
        value_template="{{value_json.grott_last_measure}}",
    ),
    # batteries
    "eacharge_today": EnergySensor("Battery charge from AC (Today)", icon="mdi:battery-arrow-up"),
    "eacCharToday": EnergySensor("Battery charge from grid today", icon="mdi:battery-arrow-up"),
    "eacharge_total": IncreasingEnergySensor("Battery charge from AC (Total)", icon="mdi:battery-arrow-up"),
    "eacCharTotal": IncreasingEnergySensor("Lifetime battery charge from grid", icon="mdi:battery-arrow-up"),
    "eacDischarToday": EnergySensor("Battery dischage today", icon="mdi:battery-arrow-down"),
    "eacDischarTotal": IncreasingEnergySensor("Lifetime battery discharge", icon="mdi:battery-arrow-down"),
    "vbat": VoltageSensor("Battery voltage"),
    "SOC": BatteryChargeSensor(
        "Battery charge",
        icon="mdi:battery-charging-60",
        # The battery charge is directly a percentage,
        # there is an error in the layout.
        value_template="{{ value_json.SOC | int }}",
    ),
    "loadpercent": PercentSensor("Load percentage"),
    "batterySoc": BatteryChargeSensor("Battery charge", icon="mdi:battery-charging-60"),
    # register 28
    "bat_Volt": VoltageSensor("Battery voltage"),
    # register 29
    "bat_dsp": VoltageSensor("Battery bus voltage"),
    "ACDischarWatt": PowerSensor("Load power"),
    "ACDischarVA": ApparentPower("Load reactive power"),
    "BatDischarWatt": PowerSensor("Battery discharge power"),
    "BatWatt": PowerSensor("Battery discharge power", icon="mdi:battery"),
    "BatDischarVA": ApparentPower("Battery discharge reactive power"),
    # taken from register 1048 of RTU manual v1.20
    "batterytype": BaseSensor(
        "Battery type",
        icon="mdi:power-settings",
        value_template="{% if value_json.batterytype == '0' %}Lithium"
        "{% elif value_json.batterytype == '1' %}Lead-acid"
        "{% elif value_json.batterytype == '2' %}Other"
        "{% else %}Unknown{% endif %}",
    ),
    "p1charge1": PowerSensor("Battery Charging Power", icon="mdi:battery-arrow-up"),
    "eharge1_tod": EnergySensor("Battery charge (Today)", icon="mdi:battery-arrow-up"),
    "eharge1_tot": IncreasingEnergySensor("Battery charge (Total)", icon="mdi:battery-arrow-up"),
    "edischarge1_tod": EnergySensor("Battery discharge (Today)", icon="mdi:battery-arrow-down"),
    "edischarge1_tot": IncreasingEnergySensor("Battery discharge (Total)", icon="mdi:battery-arrow-down"),
    "ebatDischarToday": EnergySensor("Battery discharged today", icon="mdi:battery-arrow-down"),
    "ebatDischarTotal": IncreasingEnergySensor("Lifetime battery discharged", icon="mdi:battery-arrow-down"),
    "pdischarge1": PowerSensor("Battery discharging W", icon="mdi:battery-arrow-down"),
    "ACCharCurr": CurrentSensor("Battery charging current"),
    "acchr_watt": PowerSensor("Storage charging from grid"),
    "acchr_VA": ApparentPower("Storage charging from grid reactive power"),
    "battemp": TemperatureSensor("Battery temperature", icon="mdi:thermometer"),
    "invtemp": TemperatureSensor("Inverter temperature", icon="mdi:thermometer"),
    "dcdctemp": TemperatureSensor("Battery charger temperature", icon="mdi:thermometer"),
    "spbusvolt": VoltageSensor("SP bus voltage"),
    # faults
    "faultcode": DiagnosticSensor(name="Fault code"),
    "systemfaultword1": DiagnosticSensor(name="System fault register 1"),
    "systemfaultword2": DiagnosticSensor(name="System fault register 2"),
    "systemfaultword3": DiagnosticSensor(name="System fault register 3"),
    "systemfaultword4": DiagnosticSensor(name="System fault register 4"),
    "systemfaultword5": DiagnosticSensor(name="System fault register 5"),
    "systemfaultword6": DiagnosticSensor(name="System fault register 6"),
    "systemfaultword7": DiagnosticSensor(name="System fault register 7"),
    "spdspstatus": DiagnosticSensor(name="SP DSP status"),
    "faultBit": DiagnosticSensor(name="Fault message"),
    "warningBit": DiagnosticSensor(name="Warning message"),
    "faultValue": DiagnosticSensor(name="Fault value"),
    "warningValue": DiagnosticSensor(name="Warning value"),
    "constantPowerOK": DiagnosticSensor(name="Constant power OK"),
    "systemfaultword0": DiagnosticSensor(name="System Fault Word 0"),
    "uwsysworkmode": DiagnosticSensor(name="System work mode"),
    "isof": VoltageSensor("ISO fault", icon="mdi:alert"),
    "gfcif": CurrentSensor("GFCI fault", icon="mdi:alert"),
    "dcif": CurrentSensor("DCI fault", icon="mdi:alert"),
    "vpvfault": VoltageSensor("PV voltage fault", icon="mdi:alert"),
    "vacfault": VoltageSensor("AC voltage fault", icon="mdi:alert"),
    "facfault": FrequencySensor("AC frequency fault", icon="mdi:alert"),
    "tmpfault": TemperatureSensor("Temperature fault", icon="mdi:alert"),
    # PV
    "vpv1": VoltageSensor("PV1 voltage"),
    "vpv2": VoltageSensor("PV2 voltage"),
    "ppv1": PowerSensor("PV1 Wattage"),
    "ppv2": PowerSensor("PV2 Wattage"),
    "buck1curr": CurrentSensor("Buck1 current"),
    "buck2curr": CurrentSensor("Buck2 current"),
    "op_watt": PowerSensor("Inverter active power"),
    "op_va": ApparentPower("Inverter apparent power"),
    "Inv_Curr": CurrentSensor("Inverter current"),
    "OP_Curr": CurrentSensor("Inverter consumption current"),
    "eactoday": EnergySensor("Self-Consumption (Solar + Battery) Energy - Today (eactoday)"),
    "eactotal": IncreasingEnergySensor("Self-Consumption (Solar + Battery) Energy - Total (eactotal)"),
    # temperature
    "buck1_ntc": TemperatureSensor("Buck1 temperature", icon="mdi:thermometer"),
    "buck2_ntc": TemperatureSensor("Buck2 temperature", icon="mdi:thermometer"),
    # TODO: To map
    # "nbusvolt": "",
    # "rac": "",
    # "eractoday": "",
    # "eractotal": "",
    # "plocaloadtot": "",
    # "Vac_RS": "",
    # "Vac_ST": "",
    # "Vac_TR": "",
    # "temp4": "",
    # "uwBatVolt_DSP": "",
    "voltage_l1": VoltageSensor("Phase1 Voltage"),
    "voltage_l2": VoltageSensor("Phase2 Voltage"),
    "voltage_l3": VoltageSensor("Phase3 Voltage"),
    "Current_l1": CurrentSensor("Phase1 Current"),
    "Current_l2": CurrentSensor("Phase2 Current"),
    "Current_l3": CurrentSensor("Phase3 Current"),
    "voltageL1": VoltageSensor("Phase1 Voltage"),
    "voltageL2": VoltageSensor("Phase2 Voltage"),
    "voltageL3": VoltageSensor("Phase3 Voltage"),
    "currentL1": CurrentSensor("Phase1 Current"),
    "currentL2": CurrentSensor("Phase2 Current"),
    "currentL3": CurrentSensor("Phase3 Current"),
    "L1-2_voltage": VoltageSensor("Phase1-2 Voltage"),
    "L2-3_voltage": VoltageSensor("Phase2-3 Voltage"),
    "L3-1_voltage": VoltageSensor("Phase3-1 Voltage"),
    "act_power_l1": PowerSensor("Active power L1"),
    "act_power_l2": PowerSensor("Active power L2"),
    "act_power_l3": PowerSensor("Active power L3"),
    "app_power_l1": ApparentPower("Apparent power L1"),
    "app_power_l2": ApparentPower("Apparent power L2"),
    "app_power_l3": ApparentPower("Apparent power L3"),
    "react_power_l1": ReactivePower("Reactive power L1"),
    "react_power_l2": ReactivePower("Reactive power L2"),
    "react_power_l3": ReactivePower("Reactive power L3"),
    "powerfactor_l1": PowerFactor("Power factor L1"),
    "powerfactor_l2": PowerFactor("Power factor L2"),
    "powerfactor_l3": PowerFactor("Power factor L3"),
    # "pos_rev_act_power": "",
    # "pos_act_power": "",
    # "rev_act_power": "",
    # "app_power": "",
    # "react_power": "",
    "powerfactor": PowerFactor("Power factor"),
    "power_factor": PowerFactor("Power factor"),
    # "pos_act_energy": "",
    # "rev_act_energy": "",
    # "pos_act_energy_kvar": "",
    # "rev_act_energy_kvar": "",
    # "app_energy_kvar": "",
    # "act_energy_kwh": "",
    # "react_energy_kvar": "",
    # "device": "",
    # "logstart": "",
    # "active_energy": "",
    # "reactive_energy": "",
    "activePowerL1": PowerSensor("Active power L1"),
    "activePowerL2": PowerSensor("Active power L2"),
    "activePowerL3": PowerSensor("Active power L3"),
    "reactivePowerL1": ReactivePower("Reactive power L1"),
    "reactivePowerL2": ReactivePower("Reactive power L2"),
    "reactivePowerL3": ReactivePower("Reactive power L3"),
    "apperentPowerL1": ApparentPower("Apparent power L1"),
    "apperentPowerL2": ApparentPower("Apparent power L2"),
    "apperentPowerL3": ApparentPower("Apparent power L3"),
    "powerFactorL1": PowerFactor("Power factor L1"),
    "powerFactorL2": PowerFactor("Power factor L2"),
    "powerFactorL3": PowerFactor("Power factor L3"),
    # "power": "",
    # "active_power": "",
    # "reverse_active_power": "",
    # "apparent_power": "",
    # "reactive_power": "",
    # "posiActivePower": "",
    # "reverActivePower": "",
    # "posiReactivePower": "",
    # "reverReactivePower": "",
    # "apparentEnergy": "",
    # "totalActiveEnergyL1": "",
    # "totalActiveEnergyL2": "",
    # "totalActiveEnergyL3": "",
    # "totalRectiveEnergyL1": "",
    # "totalRectiveEnergyL2": "",
    # "totalRectiveEnergyL3": "",
    # "total_energy": "",
    "l1Voltage2": VoltageSensor("Phase 1-2 Voltage"),
    "l2Voltage3": VoltageSensor("Phase 2-3 Voltage"),
    "l3Voltage1": VoltageSensor("Phase 3-1 Voltage"),
}



import pytest

from grottext.ha.constants import MQTT_HOST_CONF_KEY, MQTT_PASSWORD_CONF_KEY, MQTT_PORT_CONF_KEY, MQTT_USERNAME_CONF_KEY
from grottext.ha.ha_types import BaseSensor, DiagnosticSensor, to_dict
from grottext.ha.interface import FakeConf
from grottext.ha.mappings import mapping
from grottext.ha.mqtt import make_payload, process_conf

from .conftest import test_key, test_serial


def test_remove_none_value():
    sensor = BaseSensor(name="test")
    assert to_dict(sensor) == {"name": "test"}

    sensor = DiagnosticSensor(name="test", icon="icon_test")
    assert to_dict(sensor) == {"name": "test", "entity_category": "diagnostic", "icon": "icon_test"}


def test_to_dict():
    # test the to_dict function
    res = to_dict(
        BaseSensor(
            "Grott last data push",
            device_class="timestamp",
            value_template="{{value_json.grott_last_push}}",
        )
    )
    assert res["name"] == "Grott last data push"
    assert res["device_class"] == "timestamp"
    assert res["value_template"] == "{{value_json.grott_last_push}}"
    assert len(res.keys()) == 3
    # Even if present in the dataclass should not be serialized
    assert "unit_of_measurement" not in res


def test_manual_divider(fake_config: FakeConf):
    "Test that's the manual value template is not overwritten"
    # Alter the configuration
    value_template = "{{value_json.pvpowerout | float / 10000}}"
    key = "pvpowerout"
    mapping[key].value_template = value_template
    payload = make_payload(fake_config, test_serial, key)
    # Remove the alteration
    mapping[key].value_template = None

    assert payload["value_template"] == value_template


def test_unknown_mapping(fake_config: FakeConf):
    "Test that an unknown mapping still has a good divider"

    fake_config.recorddict[fake_config.layout].update(
        {
            "test": {"value": 290, "length": 4, "type": "num", "divide": 51},
            "test_not_num": {"value": 290, "length": 4, "type": "text", "divide": 51},
        }
    )

    # No mapping should use the raw divider
    payload = make_payload(fake_config, test_serial, "test")
    assert payload["value_template"] == "{{ value_json.test | float / 51 }}"

    # Type not text should return the raw value
    payload = make_payload(fake_config, test_serial, "test_not_num")
    assert payload["value_template"] == "{{ value_json.test_not_num }}"


def test_name_generation(fake_config):
    "Test the output of the name generation"

    # test date {"value" :76, "length" : 10, "type" : "text"},
    payload = make_payload(fake_config, test_serial, test_key)

    assert payload["name"] == "PV Output (Actual)"


def test_name_generation_non_mapped(fake_config):
    "Test the output of the name generation"

    # test date {"value" :76, "length" : 10, "type" : "text"},
    payload = make_payload(fake_config, test_serial, "duck")

    assert payload["name"] == "duck"


def test_process_conf():
    conf = FakeConf()
    try:
        res = process_conf(conf)
        pytest.fail("Should have raised an exception")
    except AttributeError:
        pass

    conf.extvar = {
        MQTT_HOST_CONF_KEY: "localhost",
        MQTT_PORT_CONF_KEY: "1883",
    }
    res = process_conf(conf)
    assert res["hostname"] == "localhost"
    assert res["port"] == 1883
    assert res["client_id"] == "Grott - HA"
    assert res["auth"] is None
    # test with auth
    conf.extvar[MQTT_USERNAME_CONF_KEY] = "user"
    conf.extvar[MQTT_PASSWORD_CONF_KEY] = "pass"
    res = process_conf(conf)
    assert res["auth"] == {"username": "user", "password": "pass"}



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from paho.mqtt.publish import multiple, single

from grottext.ha.constants import (
    MQTT_HOST_CONF_KEY,
    MQTT_PASSWORD_CONF_KEY,
    MQTT_PORT_CONF_KEY,
    MQTT_USERNAME_CONF_KEY,
)
from grottext.ha.ha_types import to_dict
from grottext.ha.interface import FakeConf
from grottext.ha.mappings import mapping

_client_name = "Grott - HA"


@dataclass
class Device:
    identifiers: List[str]
    name: str
    manufacturer: str


@dataclass
class MQTTConfigPayload:
    name: str
    unique_id: str
    state_topic: str
    device: Device
    value_template: Optional[str]
    state_class: Optional[str] = None
    device_class: Optional[str] = None
    icon: Optional[str] = None
    entity_category: Optional[str] = None
    unit_of_measurement: Optional[str] = None


def is_valid_mqtt_topic(key_name: str) -> bool:
    """Check if the key is a valid mqtt topic

    Look If the possible topic is valid, based on the MQTT spec and reserved keywords

    Parameters:
        key_name: The value of the key (e.g. "ACDischarWatt")

    Returns:
         True if the key is a valid mqtt topic, False otherwise
    """
    key_name = key_name.strip()
    # Character used to bind wildcard topics
    if key_name.startswith("#"):
        return False
    # Character used to bind single level topics
    if key_name.startswith("+"):
        return False
    # should not start or end with /
    if key_name.startswith("/"):
        return False
    if key_name.endswith("/"):
        return False
    # system topics
    if key_name.startswith("$"):
        return False
    return True


def cleanup_mqtt_values_field(values: Dict[str, Any]) -> Dict[str, Any]:
    """Cleanup the values from invalid keys

    Drop the possible invalid keys from the values and normalize the keys.

    Parameters:
        values: Original dict

    Returns:
        The cleaned-up values
    """
    return {k.strip(): v for k, v in values.items() if is_valid_mqtt_topic(k)}


def make_payload(conf: FakeConf, device: str, key: str, name: Optional[str] = None) -> Dict[str, str]:
    """Generate a MQTT payload for a sensor

    Use default values to create a sensor payload, then update with custom
    attributes if they exist.
    E.g., unit_of_measurement/total increasing/etc.

    Parameters:
        conf: The configuration object, used to extract default divider
        device: Use the device name as part of the sensor name + device
        key: The key of the sensor sent by grott
        name: The name of the sensor, if you want something different

    Returns:
        A dictionary with the MQTT configuration payload
    """

    if not device:
        msg = "device is required"
        raise AttributeError(msg)
    if not key:
        msg = "key is required"
        raise AttributeError(msg)
    if not conf.layout:
        msg = "grott config class error"
        raise AttributeError(msg)
    if conf.layout not in conf.recorddict:
        msg = "grott config class error, missing record layout"
        raise AttributeError(msg)

    if name is None:
        name = key

    sensor = mapping.get(key, None)

    value_template = None
    if sensor and sensor.value_template:
        value_template = sensor.value_template
    else:
        # Reuse the existing divide value if available and not existing
        # and apply it to the HA config
        layout = conf.recorddict[conf.layout]
        if key in layout:
            # From grottdata:207, default type is num, also process numx
            register_type = layout[key].get("type", "num")
            # The register is a numerical type
            if register_type in ("num", "numx"):
                # default divide is 1, if not found
                divider = layout[key].get("divide", "1")
                value_template = f"{{{{ value_json.{key} | float / {divider} }}}}"
            elif register_type == "log":
                # register is already a float, no need to divide
                value_template = f"{{{{ value_json.{key} | float }}}}"
            elif register_type == "logpos":
                # register is already a float, no need to divide
                value_template = f"{{{{ [value_json.{key} | float, 0.0] | max }}}}"
            elif register_type == "logneg":
                # register is already a float, no need to divide
                value_template = f"{{{{ [value_json.{key} | float, 0.0] | min }}}}"

    if value_template is None:
        value_template = f"{{{{ value_json.{key} }}}}"

    # Default configuration payload
    payload = MQTTConfigPayload(
        name="{name}",
        unique_id=f"grott_{device}_{key}",  # Generate a unique device ID
        state_topic=f"homeassistant/grott/{device}/state",
        device=Device(
            identifiers=[device],  # Group under a device
            name=device,
            manufacturer="GrowWatt",
        ),
        value_template=value_template,
    )

    if sensor is not None:
        # Update the payload with the sensor configuration
        payload.name = sensor.name
        payload.icon = sensor.icon
        payload.state_class = sensor.state_class
        payload.device_class = sensor.device_class
        payload.icon = sensor.icon
        payload.entity_category = sensor.entity_category
        payload.unit_of_measurement = sensor.unit_of_measurement

    # Generate the name of the key, with all the param available
    payload.name = payload.name.format(device=device, name=name, key=key)
    # HA automatically group the sensors if the device name is prepended

    return to_dict(payload)


def process_conf(conf: FakeConf) -> Dict[str, Union[str, int, Dict[str, Any], None]]:
    required_params = [
        MQTT_HOST_CONF_KEY,
        MQTT_PORT_CONF_KEY,
    ]
    if not all(param in conf.extvar for param in required_params):
        print("Missing configuration for ha_mqtt")
        raise AttributeError

    if MQTT_USERNAME_CONF_KEY in conf.extvar:
        auth = {
            "username": conf.extvar[MQTT_USERNAME_CONF_KEY],
            "password": conf.extvar[MQTT_PASSWORD_CONF_KEY],
        }
    else:
        auth = None

    mqtt_port = conf.extvar[MQTT_PORT_CONF_KEY]
    # Need to convert the port if passed as a string
    if isinstance(mqtt_port, str):
        port = int(mqtt_port)
    elif isinstance(mqtt_port, int):
        port = mqtt_port
    else:
        msg = f"Invalid port type: {type(mqtt_port)}"
        raise AttributeError(msg)

    return {
        "client_id": _client_name,
        "auth": auth,
        "hostname": conf.extvar[MQTT_HOST_CONF_KEY],
        "port": port,
    }


def publish_single(conf: FakeConf, topic: str, payload: str, *, retain: bool = False) -> None:
    mqtt_conf = process_conf(conf)
    single(topic, payload=payload, retain=retain, **mqtt_conf)


def publish_multiple(conf: FakeConf, msgs: List[Dict[str, Any]]) -> None:
    mqtt_conf = process_conf(conf)
    multiple(msgs, **mqtt_conf)



import pytest

from grottext.ha.interface import FakeConf
from grottext.ha.mqtt import cleanup_mqtt_values_field, is_valid_mqtt_topic, make_payload


def test_is_valid_mqtt_topic():
    assert is_valid_mqtt_topic("plocaloadr") is True
    assert is_valid_mqtt_topic("#nbusvolt") is False
    assert is_valid_mqtt_topic("/test") is False
    assert is_valid_mqtt_topic("test/") is False
    assert is_valid_mqtt_topic("+test") is False
    assert is_valid_mqtt_topic("$test") is False  # System topic


def test_cleanup_values():
    values = {
        "plocaloadr": 1,
        "epv1today ": 32,
        "#nbusvolt": 2,
        "/test": 3,
        "test/": 4,
        "+test": 5,
        "$test": 6,
    }
    assert cleanup_mqtt_values_field(values) == {
        "plocaloadr": 1,
        "epv1today": 32,
    }


def test_unknown_layout(fake_config: FakeConf):
    "Test that's the manual value template is not overwritten"
    # Alter the configuration
    fake_config.layout = "FAILURE"
    with pytest.raises(AttributeError, match="grott config class error, missing record layout"):
        make_payload(fake_config, "NCO7410", "pvpowerout")



CONFIG_TOPIC = "homeassistant/{sensor_type}/grott/{device}_{attribut}/config"
STATE_TOPIC = "homeassistant/grott/{device}/state"

MQTT_HOST_CONF_KEY = "ha_mqtt_host"
MQTT_PORT_CONF_KEY = "ha_mqtt_port"
MQTT_USERNAME_CONF_KEY = "ha_mqtt_user"
MQTT_PASSWORD_CONF_KEY = "ha_mqtt_password"
MQTT_RETAIN_CONF_KEY = "ha_mqtt_retain"



# Used to simulate a Conf object from grott
from typing import Any, Dict, Union


class FakeConf:
    """A fake configuration object imitating the one from grott.

    !!! danger "This is a fake object"
        Only used for testing.
    """

    def __init__(self) -> None:
        self.recorddict: Dict[str, Any] = {}
        self.layout = ""
        self.extvar: Dict[str, Union[str, int, bool]] = {}
        self.verbose = True



from dataclasses import asdict, dataclass
from typing import Any, Dict, Optional


@dataclass
class BaseSensor:
    name: str
    icon: Optional[str] = None
    value_template: Optional[str] = None
    state_class: Optional[str] = None
    device_class: Optional[str] = None
    unit_of_measurement: Optional[str] = None
    entity_category: Optional[str] = None


@dataclass
class DiagnosticSensor(BaseSensor):
    entity_category: str = "diagnostic"
    icon = "mdi:information-outline"


@dataclass
class MeasurementSensor(BaseSensor):
    state_class: str = "measurement"
    device_class: Optional[str] = None
    unit_of_measurement: Optional[str] = None


@dataclass
class PercentSensor(MeasurementSensor):
    unit_of_measurement: str = "%"


@dataclass
class VoltageSensor(MeasurementSensor):
    device_class: str = "voltage"
    unit_of_measurement: str = "V"


@dataclass
class CurrentSensor(MeasurementSensor):
    device_class: str = "current"
    unit_of_measurement: str = "A"


@dataclass
class PowerSensor(MeasurementSensor):
    device_class: str = "power"
    unit_of_measurement: str = "W"


@dataclass
class FrequencySensor(MeasurementSensor):
    device_class: str = "frequency"
    unit_of_measurement: str = "Hz"


@dataclass
class TemperatureSensor(MeasurementSensor):
    device_class: str = "temperature"
    unit_of_measurement: str = "°C"


@dataclass
class DurationSensor(MeasurementSensor):
    device_class: str = "duration"
    unit_of_measurement: str = "h"


@dataclass
class ApparentPower(MeasurementSensor):
    device_class: str = "apparent_power"
    unit_of_measurement: str = "VA"


@dataclass
class ReactivePower(MeasurementSensor):
    device_class: str = "reactive_power"
    unit_of_measurement: str = "var"


@dataclass
class BatteryChargeSensor(MeasurementSensor):
    device_class: str = "battery"
    unit_of_measurement: str = "%"


@dataclass
class EnergySensor(MeasurementSensor):
    state_class: str = "total"
    device_class: str = "energy"
    unit_of_measurement: str = "kWh"


@dataclass
class IncreasingEnergySensor(MeasurementSensor):
    state_class: str = "total_increasing"
    device_class: str = "energy"
    unit_of_measurement: str = "kWh"


@dataclass
class PowerFactor(MeasurementSensor):
    device_class: str = "power_factor"


def to_dict(obj: Any) -> Dict[str, Any]:
    """Convert a dataclass object to dict

    Parameters:
        obj: The sensor object to convert

    Returns:
        A dictionary representation of the object
    """
    dict_obj = asdict(obj)
    # Remove None values
    return {k: v for k, v in dict_obj.items() if v is not None}
